import { ReactNode } from 'react';
export interface IVarUIProps<T extends object> {
    /**
     * A JavaScript object or array to be mutated by the input components.
     */
    values: T;
    /**
     * A JavaScript object holding error information.
     */
    errors?: any;
    /**
     * @deprecated Replaced by onChange
     */
    updateValues?: (values: T) => void;
    /**
     * The function to be called with the entire changed object.
     */
    onChange?: (values: T) => void;
    /**
     * The function to be called when one value is changed.
     */
    onChangeValue?: (path: string, newValue: any) => void;
    /**
     * Additional class names for the wrapper object.
     */
    className?: string;
    /**
     * Input components (or any other children).
     */
    children?: ReactNode;
}
/**
 * This is the main component which provides a Context for other components.
 * It is not required to use this component - other components accept
 * `onChange` and `value` properties which provide a similar functionality.
 */
export declare const VarUI: <T extends object>(props: IVarUIProps<T>) => JSX.Element;
